use crate::model::SectionRef;
use crate::section_interface::SectionInterface;
use mpsc::error::SendError;
use std::{
    collections::VecDeque,
    mem::swap,
    sync::{
        atomic::{AtomicI32, Ordering},
        Arc,
    },
    time::Duration,
};
use thiserror::Error;
use tokio::{
    spawn,
    sync::mpsc,
    time::{delay_for, Instant},
};
use tracing::{debug, trace, trace_span};

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct RunHandle(i32);

#[derive(Debug)]
struct SectionRunnerInner {
    next_run_id: AtomicI32,
}

impl SectionRunnerInner {
    fn new() -> Self {
        Self {
            next_run_id: AtomicI32::new(1),
        }
    }
}

#[derive(Clone, Debug)]
enum RunnerMsg {
    Quit,
    QueueRun(RunHandle, SectionRef, Duration),
    CancelRun(RunHandle),
    CancelAll,
}

#[derive(Clone, Debug, PartialEq)]
enum RunState {
    Waiting,
    Running { start_time: Instant },
    Finished,
    Cancelled,
}

#[derive(Debug)]
struct SecRun {
    handle: RunHandle,
    section: SectionRef,
    duration: Duration,
    state: RunState,
}

impl SecRun {
    fn start(&mut self, interface: &dyn SectionInterface) {
        use RunState::*;
        debug!(section_id = self.section.id, "starting running section");
        interface.set_section_state(self.section.interface_id, true);
        self.state = Running {
            start_time: Instant::now(),
        };
    }

    fn is_running(&self) -> bool {
        matches!(self.state, RunState::Running{..})
    }

    fn finish(&mut self, interface: &dyn SectionInterface) {
        if self.is_running() {
            debug!(section_id = self.section.id, "finished running section");
            interface.set_section_state(self.section.interface_id, false);
            self.state = RunState::Finished;
        }
    }

    fn cancel(&mut self, interface: &dyn SectionInterface) {
        if self.is_running() {
            debug!(section_id = self.section.id, "cancelling running section");
            interface.set_section_state(self.section.interface_id, false);
        }
        self.state = RunState::Cancelled;
    }
}

mod option_future {
    use pin_project::pin_project;
    use std::{
        future::Future,
        ops::Deref,
        pin::Pin,
        task::{Context, Poll},
    };

    #[pin_project]
    #[derive(Debug, Clone)]
    #[must_use = "futures do nothing unless you `.await` or poll them"]
    pub struct OptionFuture<F>(#[pin] Option<F>);

    impl<F: Future> Future for OptionFuture<F> {
        type Output = Option<F::Output>;

        fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
            match self.project().0.as_pin_mut() {
                Some(x) => x.poll(cx).map(Some),
                None => Poll::Ready(None),
            }
        }
    }

    impl<F> Deref for OptionFuture<F> {
        type Target = Option<F>;

        fn deref(&self) -> &Self::Target {
            &self.0
        }
    }

    impl<T> From<Option<T>> for OptionFuture<T> {
        fn from(option: Option<T>) -> Self {
            OptionFuture(option)
        }
    }
}

use option_future::OptionFuture;

async fn runner_task(
    interface: Arc<dyn SectionInterface + Sync>,
    mut msg_recv: mpsc::Receiver<RunnerMsg>,
) {
    use RunState::*;

    let span = trace_span!("runner_task");
    let _enter = span.enter();

    let mut running = true;
    let mut run_queue: VecDeque<SecRun> = VecDeque::new();
    let mut delay_future: OptionFuture<_> = None.into();
    while running {
        if let Some(current_run) = run_queue.front_mut() {
            let done = match current_run.state {
                Waiting => {
                    current_run.start(&*interface);
                    delay_future = Some(delay_for(current_run.duration)).into();
                    false
                }
                Running { start_time } => {
                    if start_time.elapsed() >= current_run.duration {
                        current_run.finish(&*interface);
                        delay_future = None.into();
                        true
                    } else {
                        false
                    }
                }
                Cancelled | Finished => true,
            };

            if done {
                run_queue.pop_front();
                continue;
            }
        }

        let mut handle_msg = |msg: Option<RunnerMsg>| {
            let msg = msg.expect("SectionRunner channel closed");
            use RunnerMsg::*;
            trace!(msg = debug(&msg), "runner_task recv");
            match msg {
                Quit => running = false,
                QueueRun(handle, section, duration) => {
                    run_queue.push_back(SecRun {
                        handle,
                        section,
                        duration,
                        state: Waiting,
                    });
                }
                CancelRun(handle) => {
                    for run in &mut run_queue {
                        if run.handle != handle {
                            continue;
                        }
                        trace!(handle = handle.0, "cancelling run by handle");
                        run.cancel(&*interface);
                    }
                }
                CancelAll => {
                    let mut old_runs = VecDeque::new();
                    swap(&mut old_runs, &mut run_queue);
                    trace!(count = old_runs.len(), "cancelling all runs");
                    for mut run in old_runs {
                        run.cancel(&*interface);
                    }
                }
            }
        };
        let delay_done = || {
            trace!("delay done");
        };
        tokio::select! {
            msg = msg_recv.recv() => handle_msg(msg),
            _ = &mut delay_future, if delay_future.is_some() => delay_done()
        };
    }
}

#[derive(Debug, Clone, Error)]
#[error("the SectionRunner channel is closed")]
pub struct ChannelClosed;

pub type Result<T, E = ChannelClosed> = std::result::Result<T, E>;

impl<T> From<SendError<T>> for ChannelClosed {
    fn from(_: SendError<T>) -> Self {
        Self
    }
}

#[derive(Clone, Debug)]
pub struct SectionRunner {
    inner: Arc<SectionRunnerInner>,
    msg_send: mpsc::Sender<RunnerMsg>,
}

#[allow(dead_code)]
impl SectionRunner {
    pub fn new(interface: Arc<dyn SectionInterface + Sync>) -> Self {
        let (msg_send, msg_recv) = mpsc::channel(8);
        spawn(runner_task(interface, msg_recv));
        Self {
            inner: Arc::new(SectionRunnerInner::new()),
            msg_send,
        }
    }

    pub async fn quit(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Quit).await?;
        Ok(())
    }

    pub async fn queue_run(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> Result<RunHandle> {
        let run_id = self.inner.next_run_id.fetch_add(1, Ordering::Relaxed);
        let handle = RunHandle(run_id);
        self.msg_send
            .send(RunnerMsg::QueueRun(handle.clone(), section, duration))
            .await?;
        Ok(handle)
    }

    pub async fn cancel_run(&mut self, handle: RunHandle) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelRun(handle)).await?;
        Ok(())
    }

    pub async fn cancel_all(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelAll).await?;
        Ok(())
    }

    pub async fn pause(&mut self) -> Result<()> {
        todo!()
    }

    pub async fn unpause(&mut self) -> Result<()> {
        todo!()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::section_interface::MockSectionInterface;
    use crate::{
        model::Section,
        trace_listeners::{EventListener, Filters, SpanFilters, SpanListener},
    };
    use tokio::time::{pause, resume};
    use tracing_subscriber::prelude::*;

    #[tokio::test]
    async fn test_quit() {
        let quit_msg = EventListener::new(
            Filters::new()
                .target("sprinklers_rs::section_runner")
                .message("runner_task recv")
                .field_value("msg", "Quit"),
        );
        let task_span = SpanListener::new(
            SpanFilters::new()
                .target("sprinklers_rs::section_runner")
                .name("runner_task"),
        );
        let subscriber = tracing_subscriber::registry()
            .with(quit_msg.clone())
            .with(task_span.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        let interface = MockSectionInterface::new(6);
        let mut runner = SectionRunner::new(Arc::new(interface));
        tokio::task::yield_now().await;
        runner.quit().await.unwrap();
        tokio::task::yield_now().await;

        assert_eq!(quit_msg.get_count(), 1);
        assert_eq!(task_span.get_exit_count(), 1);
    }

    fn make_sections_and_interface() -> (Vec<SectionRef>, Arc<MockSectionInterface>) {
        let interface = Arc::new(MockSectionInterface::new(2));
        let sections: Vec<SectionRef> = vec![
            Arc::new(Section {
                id: 1,
                name: "Section 1".into(),
                interface_id: 0,
            }),
            Arc::new(Section {
                id: 2,
                name: "Section 2".into(),
                interface_id: 1,
            }),
        ];
        (sections, interface)
    }

    fn assert_section_states(interface: &MockSectionInterface, states: &[bool]) {
        for (id, state) in states.iter().enumerate() {
            assert_eq!(
                interface.get_section_state(id as u32),
                *state,
                "section interface id {} did not match",
                id
            );
        }
    }

    async fn advance(dur: Duration) {
        // HACK: advance should really be enough, but we need another yield_now
        tokio::time::advance(dur).await;
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_queue() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        assert_section_states(&interface, &[false, false]);

        // Queue single section, make sure it runs
        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        pause();
        advance(Duration::from_secs(1)).await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        // Queue two sections, make sure they run one at a time
        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        advance(Duration::from_secs(1)).await;

        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        resume();

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_cancel_run() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let run1 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run2 = runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        pause();

        advance(Duration::from_secs(1)).await;

        assert_section_states(&interface, &[false, true]);

        runner.cancel_run(run1).await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        runner.cancel_run(run3).await.unwrap();
        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        resume();

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }

    #[tokio::test]
    async fn test_cancel_all() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[0].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        pause();

        advance(Duration::from_secs(1)).await;

        assert_section_states(&interface, &[false, true]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, false]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, false]);

        resume();

        runner.quit().await.unwrap();
        tokio::task::yield_now().await;
    }
}
